"""
Click parameter types
"""

import shlex

import click
from click.exceptions import BadParameter

from . import composer


class ComposerPackages(click.ParamType):
    """
    Click type to represent a list of composer packages
    """

    name = "composer_packages"

    def convert(self, value, param, ctx):
        packages = shlex.split(value)

        return list(map(ComposerPackage.parse_package, packages))


class ComposerPackage(click.ParamType):
    """
    Click type to represent a single composer package
    """

    name = "composer_package"

    def convert(self, value, param, ctx):
        return ComposerPackage.parse_package(value)

    @staticmethod
    def parse_package(value):
        """
        Parse a composer package version string into a ComposerVersion type
        """
        try:
            composer_version = composer.ComposerVersion.from_package_string(value)
        except ValueError as err:
            raise BadParameter from err
        return composer_version


COMPOSER_PACKAGES = ComposerPackages()
COMPOSER_PACKAGE = ComposerPackage()



"""
Composer support functions
"""

import json
import pkgutil
import shutil
import subprocess
import sys

from . import util


def is_present():
    """
    Check if composer is present
    """
    return shutil.which("composer") is not None


def get_version():
    """
    Attempt to get composer version
    """
    if not is_present():
        raise RuntimeError("Composer not installed")

    composer_run = subprocess.run(
        "composer --version", shell=True, check=False, capture_output=True
    )

    if not composer_run.stdout.startswith(b"Composer version "):
        raise RuntimeError("Unexpected output from composer --version")

    version_string = composer_run.stdout
    end_index = version_string.find(b" ", 17)
    version = version_string[17:end_index]

    return version.decode("utf-8")


def is_version_2():
    """
    Check if we have composer 2 installed
    """
    version = get_version()
    return version[0] == "2"


def run_install():
    """
    Run composer install and handle errors
    """
    if not is_present():
        util.write_error("Cannot find composer. Aborting...")
        sys.exit(4)

    composer_run = subprocess.run(
        "composer install --no-interaction --ignore-platform-reqs -o",
        shell=True,
        check=False,
    )
    if composer_run.returncode != 0:
        util.write_error("Error when running 'composer install'. Aborting...")
        util.write_error("Make sure you have set an adequate PHP memory limit.")
        util.write_warning(
            "Read https://getcomposer.org/doc/articles/troubleshooting.md#memory-limit-errors"
            + " for more details"
        )
        sys.exit(4)


def require_packages(packages, dev=False):
    """
    Run composer require or require --dev
    """
    if not is_present():
        util.write_error("Cannot find composer. Aborting...")
        sys.exit(4)

    if len(packages) == 0:
        return

    cmd = "composer require "
    if dev:
        cmd += "--dev "
    cmd += " ".join([str(p) for p in packages])

    composer_run = subprocess.run(cmd, shell=True, check=False)
    if composer_run.returncode != 0:
        util.write_error("Error when running 'composer require'. Aborting...")
        util.write_error("Make sure you have set an adequate PHP memory limit.")
        util.write_warning(
            "Read https://getcomposer.org/doc/articles/troubleshooting.md#memory-limit-errors"
            + " for more details"
        )
        sys.exit(4)


def get_drupal_template(name, description, core, core_version, docroot, cache_service):
    """
    Get the composer template from package and modify it as per given options
    """
    composer = json.loads(pkgutil.get_data(__name__, "files/drupal/composer.json"))
    composer["name"] = name
    composer["description"] = description
    composer["extra"]["drupal-scaffold"]["locations"]["web-root"] = docroot + "/"
    composer["extra"]["installer-paths"] = {
        docroot + "/core": ["type:drupal-core"],
        docroot + "/libraries/{$name}": ["type:drupal-library"],
        docroot + "/modules/contrib/{$name}": ["type:drupal-module"],
        docroot + "/profiles/contrib/{$name}": ["type:drupal-profile"],
        docroot + "/themes/contrib/{$name}": ["type:drupal-theme"],
        "drush/Commands/contrib/{$name}": ["type:drupal-drush"],
        docroot + "/modules/custom/{$name}": ["type:drupal-custom-module"],
        docroot + "/themes/custom/{$name}": ["type:drupal-custom-theme"],
    }

    if core == "core":
        composer["require"]["drupal/core"] = core_version
    if core == "recommended":
        composer["require"]["drupal/core-recommended"] = core_version
    if cache_service == "redis":
        composer["require"]["drupal/redis"] = "^1.4"
    if cache_service == "memcache":
        composer["require"]["drupal/memcache"] = "^2.0"
    composer["require"]["drupal/core-composer-scaffold"] = core_version

    if not is_version_2():
        composer["require"]["zaporylie/composer-drupal-optimizations"] = "^1.1"

    return composer


def sort_packages(composer):
    """
    Sort the packages in a composer array
    """
    composer["require"] = util.sort_dictionary_by_keys(composer["require"])
    composer["require-dev"] = util.sort_dictionary_by_keys(composer["require-dev"])
    return composer


class ComposerVersion:
    """
    Utility class to represent a composer package version
    """

    def __init__(self, name, version=""):
        parts = name.split("/")
        if len(parts) != 2:
            raise ValueError(f"{name} is not a valid composer package name")
        self.name = name
        self.package_vendor = parts[0]
        self.package_name = parts[1]

        if " " in version and not ("'" in version or '"' in version):
            version = f'"{version}"'
        self.version = version

    @staticmethod
    def from_package_string(package_string):
        """
        Parse a composer package version string into a ComposerVersion type
        """
        if ":" in package_string:
            name, version = package_string.split(":")
        else:
            name = package_string
            version = ""

        parts = name.split("/")
        if len(parts) == 1:
            name = "drupal/" + name
        elif len(parts) != 2:
            raise ValueError(f"{package_string} is not a valid composer package name")
        return ComposerVersion(name, version)

    def get_name(self):
        """
        Get the name of the composer package
        """
        return self.name

    def get_package_vendor(self):
        """
        Get the name of the package vendor, e.g., "drupal" in "drupal/module"
        """
        return self.package_vendor

    def get_package_name(self):
        """
        Get the name of the package name, e.g., "module" in "drupal/module"
        """
        return self.package_name

    def get_version(self):
        """
        Get the version of the composer package (if any)
        """
        return self.version

    def __str__(self) -> str:
        """
        Get string representation of the composer string
        """
        package_str = self.name
        if self.version:
            package_str += f':"{self.version}"'
        return package_str

    def __repr__(self) -> str:
        """
        Get string representation of the composer string
        """
        return self.__str__()



"""
Add Lando support to a Drupal codebase
"""

import json
import os

from . import util
from . import drupal


def main():
    """
    Main entrypoint for init-lando
    """
    if not os.path.exists("composer.json"):
        util.write_error("Could not find composer.json in the current directory")
        return 2

    composer = json.loads(util.read_file("composer.json"))
    name = composer["name"].split("/")
    name = name[1] if len(name) == 2 else name[0]
    try:
        scaffold_opts = composer["extra"]["drupal-scaffold"]
        docroot = scaffold_opts["locations"]["web-root"].strip("/")
    except KeyError:
        docroot = "." if composer["name"] == "drupal/drupal" else ""

    if docroot == "":
        util.write_error(
            "Could not determine docroot. Make sure your composer.json is valid."
        )
        return 3

    cache = ""
    if "drupal/redis" in composer["require"].keys():
        cache = "redis"
    elif "drupal/memcache" in composer["require"].keys():
        cache = "memcached"

    generate_lando_files(name, docroot, cache)
    return 0


def generate_lando_files(name, docroot, cache):
    """
    Generate Lando files in the docroot
    """
    services = ""
    tooling = ""
    if cache == "redis":
        services = """  cache:
    app_mount: disabled
    type: redis:5
"""
        tooling = """  redis-cli:
    service: cache
"""
    elif cache == "memcached":
        services = """  cache:
    app_mount: disabled
    type: memcached:1
"""

    yml = util.read_package_file("files/lando/lando.yml")
    yml = yml.replace("{name}", name)
    yml = yml.replace("{docroot}", docroot)
    yml = yml.replace("{services}", services)
    yml = yml.replace("{tooling}", tooling)
    util.write_file(".lando.yml", yml)

    os.makedirs(".lando", mode=0o755, exist_ok=True)
    util.copy_package_file("files/lando/php.ini", ".lando/php.ini")

    # Generate lando development override configuration.
    dir_default = f"{docroot}/sites/default"
    if not os.path.exists(dir_default):
        util.write_error(
            f'The "{dir_default}" directory is missing. '
            + "Unable to generate lando override configuration files."
        )
        util.write_info(
            "This is probably due to composer installation failure. "
            + "Run init-lando after running composer install."
        )
        return 2

    lando_settings = util.read_package_file("files/lando/settings.lando.php")
    if cache == "redis":
        lando_settings += util.read_package_file("files/lando/lando.redis.php")
    elif cache == "memcached":
        lando_settings += util.read_package_file("files/lando/lando.memcache.php")
    util.write_file(docroot + "/sites/default/settings.lando.php", lando_settings)

    settings_file = drupal.ensure_settings_file(docroot)
    drupal.modify_settings_file(
        settings_file, "include $app_root . '/' . $site_path . '/settings.lando.php';"
    )

    util.write_info("Successfully created Lando configuration files.")
    util.write_important(
        "Change the database image from mariadb to your desired image in .lando.yml."
    )

    return 0



"""
Add Gitlab support to a Drupal codebase
"""

import json
import os

from . import util
from . import drupal


def main():
    """
    Main entrypoint for init-gitlab
    """
    if not os.path.exists("composer.json"):
        util.write_error("Could not find composer.json in the current directory")
        return 2

    composer = json.loads(util.read_file("composer.json"))
    try:
        scaffold_opts = composer["extra"]["drupal-scaffold"]
        docroot = scaffold_opts["locations"]["web-root"].strip("/")
    except KeyError:
        docroot = "." if composer["name"] == "drupal/drupal" else ""

    if docroot == "":
        util.write_error(
            "Could not determine docroot. Make sure your composer.json is valid."
        )
        return 3

    generate_gitlab_files(docroot)
    return 0


def generate_gitlab_files(docroot):
    """
    Generate gitlab files in the docroot
    """

    yml = util.read_package_file("files/gitlab/gitlab-ci.yml")
    yml = yml.replace("{_docroot_}", docroot)
    util.write_file(".gitlab-ci.yml", yml)

    os.makedirs(".gitlab", mode=0o755, exist_ok=True)
    util.copy_package_file(
        "files/gitlab/settings.local.php", ".gitlab/settings.local.php"
    )
    ci_sh = util.read_package_file("files/gitlab/ci.sh")
    ci_sh = ci_sh.replace("{_docroot_}", docroot)
    util.write_file(".gitlab/ci.sh", ci_sh)
    os.chmod(".gitlab/ci.sh", 0o755)

    # Generate gitlab development override configuration.
    dir_default = f"{docroot}/sites/default"
    if not os.path.exists(dir_default):
        util.write_error(
            f'The "{dir_default}" directory is missing. '
            + "Unable to generate GitLab CI configuration files."
        )
        util.write_info(
            "This is probably due to composer installation failure. "
            + "Run init-gitlab after running composer install."
        )
        return 2

    settings_file = drupal.ensure_settings_file(docroot)
    drupal.modify_settings_file(
        settings_file,
        """if (file_exists($app_root . '/' . $site_path . '/settings.local.php')) {
  include $app_root . '/' . $site_path . '/settings.local.php';
}""",
    )

    util.write_info("Successfully created GitLab CI files.")
    util.write_important(
        "Change the database image from mariadb to your desired image in .gitlab-ci.yml."
    )

    return 0
